from .base import BaseLearner
from .ensemble import EnsembleLearner
from .few_shot_svm import FewShotSVMLearner
from .matching import MatchingLearner
from .ner import NERLearner
from .object_detection import ObjectDetectionLearner
from .semantic_segmentation import SemanticSegmentationLearner
