_base_ = "./dino-5scale_swin-l_8xb2-12e_coco.py"
max_epochs = 36
train_cfg = dict(type="EpochBasedTrainLoop", max_epochs=max_epochs, val_interval=1)
param_scheduler = [dict(type="MultiStepLR", begin=0, end=max_epochs, by_epoch=True, milestones=[27, 33], gamma=0.1)]
